import numpy as np
import pyopencl as cl
import pyopencl.clmath  # noqa

from meshmode.discretization import Discretization
from meshmode.discretization.poly_element import \
        InterpolatoryQuadratureSimplexGroupFactory

from pytential import bind, sym, norm  # noqa

# {{{ set some constants for use below

nelements = 50
mesh_order = 4
target_order = 8
ovsmp_target_order = 4*target_order
qbx_order = 4
fmm_order = False
mu = 1

# }}}


def main():
    import logging
    logging.basicConfig(level=logging.INFO)

    cl_ctx = cl.create_some_context()
    queue = cl.CommandQueue(cl_ctx)

    from meshmode.mesh.generation import (  # noqa
            make_curve_mesh, starfish, ellipse, drop)
    mesh = make_curve_mesh(
            lambda t: ellipse(1, t),
            np.linspace(0, 1, nelements+1),
            target_order)
    coarse_density_discr = Discretization(
            cl_ctx, mesh,
            InterpolatoryQuadratureSimplexGroupFactory(target_order))

    from pytential.qbx import QBXLayerPotentialSource
    stick_out = 0.05
    qbx, _ = QBXLayerPotentialSource(
            coarse_density_discr, fine_order=ovsmp_target_order, qbx_order=qbx_order,
            fmm_order=fmm_order, target_stick_out_factor=stick_out
            ).with_refinement()

    density_discr = qbx.density_discr
    mv_normal = bind(density_discr, sym.normal(2))(queue)
    normal = mv_normal.as_vector()
    path_length = bind(density_discr, sym.integral(2, 1, 1))(queue)

    # {{{ describe bvp

#    from pytential.symbolic.stokes import StressletWrapper, StokesletWrapper
    from stokes import StressletWrapper, StokesletWrapper
    from pytools.obj_array import make_obj_array
    dim=2
    cse = sym.cse

    sigma_sym = sym.make_sym_vector("sigma", dim)
    sqrt_w = sym.sqrt_jac_q_weight(2)
    inv_sqrt_w_sigma = cse(sigma_sym/sqrt_w)
    mean_inv_sqrt_w_sigma = sym.ones_vec(dim=2) * sym.mean(2, 1, inv_sqrt_w_sigma)
    int_inv_sqrt_w_sigma = sym.ones_vec(dim=2) * sym.integral(2, 1, inv_sqrt_w_sigma)
    nvec_sym = sym.make_sym_vector("normal", dim)
    mu_sym = sym.var("mu")

    # -1 for interior Dirichlet
    # +1 for exterior Dirichlet
    loc_sign = 1

    stresslet_obj = StressletWrapper(dim=2)
    stokeslet_obj = StokesletWrapper(dim=2)
    bdry_op_sym = -loc_sign * 0.5 * sigma_sym - sqrt_w * (
                            stresslet_obj.apply(inv_sqrt_w_sigma, nvec_sym, mu_sym, qbx_forced_limit='avg') +
                            stokeslet_obj.apply(inv_sqrt_w_sigma - mean_inv_sqrt_w_sigma, mu_sym, qbx_forced_limit='avg') + 
                            (0.5/np.pi) * int_inv_sqrt_w_sigma )

    # }}}

    bound_op = bind(qbx, bdry_op_sym)

    # {{{ fix rhs and solve

    def fund_soln(x, y, loc, strength):
        #with direction (1,0) for point source
        r = cl.clmath.sqrt((x - loc[0])**2 + (y - loc[1])**2)
        scaling = strength/(4*np.pi*mu)
        xcomp = (-cl.clmath.log(r) + (x - loc[0])**2/r**2) * scaling
        ycomp = ((x - loc[0])*(y - loc[1])/r**2) * scaling
        return [ xcomp, ycomp ]

    def rotlet_soln(x, y, loc):
        r = cl.clmath.sqrt((x - loc[0])**2 + (y - loc[1])**2)
        xcomp = -(y - loc[1])/r**2
        ycomp = (x - loc[0])/r**2
        return [ xcomp, ycomp ] 

    def fund_and_rot_soln(x, y, loc, strength):
        #with direction (1,0) for point source
        r = cl.clmath.sqrt((x - loc[0])**2 + (y - loc[1])**2)
        scaling = strength/(4*np.pi*mu)
        xcomp = (-cl.clmath.log(r) + (x - loc[0])**2/r**2) * scaling - (y - loc[1])*strength*0.125/r**2
        ycomp = ((x - loc[0])*(y - loc[1])/r**2) * scaling + (x - loc[0])*strength*0.125/r**2
        return [ xcomp, ycomp ]
       

    nodes = density_discr.nodes().with_queue(queue)
    fund_soln_loc = np.array([0.5, -0.2])
    strength = 100.
    bc = fund_and_rot_soln(nodes[0], nodes[1], fund_soln_loc, strength)
 
    omega_sym = sym.make_sym_vector("omega", dim)
    u_A_sym = stokeslet_obj.apply(omega_sym, mu_sym, qbx_forced_limit=1)

    omega = [cl.array.to_device(queue, (strength/path_length)*np.ones(len(nodes[0]))), cl.array.to_device(queue, np.zeros(len(nodes[0])))]
    bvp_rhs = bind(qbx, sqrt_w*(sym.make_sym_vector("bc",dim) - u_A_sym))(queue, bc=bc, mu=mu, omega=omega)

    from pytential.solve import gmres
    gmres_result = gmres(
             bound_op.scipy_op(queue, "sigma", np.float64, mu=mu, normal=normal),
             bvp_rhs, tol=1e-9, progress=True,
             stall_iterations=0,
             hard_failure=True)

    # }}}



if __name__ == "__main__":
    main()
